from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Awaitable
from dataclasses import dataclass, field
import json
import random
from time import perf_counter
from typing import final

from mcstatus.address import Address
from mcstatus.protocol.connection import Connection, TCPAsyncSocketConnection, TCPSocketConnection
from mcstatus.responses import JavaStatusResponse, RawJavaResponse


@dataclass
class _BaseServerPinger(ABC):
    connection: TCPSocketConnection | TCPAsyncSocketConnection
    address: Address
    version: int = 47
    ping_token: int = field(default_factory=lambda: random.randint(0, (1 << 63) - 1))

    def handshake(self) -> None:
        """Writes the initial handshake packet to the connection."""
        packet = Connection()
        packet.write_varint(0)
        packet.write_varint(self.version)
        packet.write_utf(self.address.host)
        packet.write_ushort(self.address.port)
        packet.write_varint(1)  # Intention to query status

        self.connection.write_buffer(packet)

    @abstractmethod
    def read_status(self) -> JavaStatusResponse | Awaitable[JavaStatusResponse]:
        """Make a status request and parse the response."""
        raise NotImplementedError

    @abstractmethod
    def test_ping(self) -> float | Awaitable[float]:
        """Send a ping token and measure the latency."""
        raise NotImplementedError

    def _handle_status_response(self, response: Connection, start: float, end: float) -> JavaStatusResponse:
        """Given a response buffer (already read from connection), parse and build the JavaStatusResponse."""
        if response.read_varint() != 0:
            raise IOError("Received invalid status response packet.")
        try:
            raw: RawJavaResponse = json.loads(response.read_utf())
        except ValueError:
            raise IOError("Received invalid JSON")

        try:
            latency_ms = (end - start) * 1000
            return JavaStatusResponse.build(raw, latency=latency_ms)
        except KeyError as e:
            raise IOError(f"Received invalid status response: {e!r}")

    def _handle_ping_response(self, response: Connection, start: float, end: float) -> float:
        """Given a ping response buffer, validate token and compute latency."""
        if response.read_varint() != 1:
            raise IOError("Received invalid ping response packet.")
        received_token = response.read_long()
        if received_token != self.ping_token:
            raise IOError(f"Received mangled ping response (expected token {self.ping_token}, got {received_token})")
        return (end - start) * 1000


@final
@dataclass
class ServerPinger(_BaseServerPinger):
    connection: TCPSocketConnection  # pyright: ignore[reportIncompatibleVariableOverride]

    def read_status(self) -> JavaStatusResponse:
        """Send the status request and read the response."""
        request = Connection()
        request.write_varint(0)  # Request status
        self.connection.write_buffer(request)

        start = perf_counter()
        response = self.connection.read_buffer()
        end = perf_counter()
        return self._handle_status_response(response, start, end)

    def test_ping(self) -> float:
        """Send a ping token and measure the latency."""
        request = Connection()
        request.write_varint(1)  # Test ping
        request.write_long(self.ping_token)
        start = perf_counter()
        self.connection.write_buffer(request)

        response = self.connection.read_buffer()
        end = perf_counter()
        return self._handle_ping_response(response, start, end)


@final
@dataclass
class AsyncServerPinger(_BaseServerPinger):
    connection: TCPAsyncSocketConnection  # pyright: ignore[reportIncompatibleVariableOverride]

    async def read_status(self) -> JavaStatusResponse:
        """Send the status request and read the response."""
        request = Connection()
        request.write_varint(0)  # Request status
        self.connection.write_buffer(request)

        start = perf_counter()
        response = await self.connection.read_buffer()
        end = perf_counter()
        return self._handle_status_response(response, start, end)

    async def test_ping(self) -> float:
        """Send a ping token and measure the latency."""
        request = Connection()
        request.write_varint(1)  # Test ping
        request.write_long(self.ping_token)
        start = perf_counter()
        self.connection.write_buffer(request)

        response = await self.connection.read_buffer()
        end = perf_counter()
        return self._handle_ping_response(response, start, end)



import sys
import time
from unittest import mock

import pytest

from mcstatus.address import Address
from mcstatus.bedrock_status import BedrockServerStatus
from mcstatus.responses import BedrockStatusResponse


def test_bedrock_response_is_expected_type():
    data = (
        b"\x1c\x00\x00\x00\x00\x00\x00\x00\x004GT\x00\xb8\x83D\xde\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd"
        b"\x124Vx\x00wMCPE;\xc2\xa7r\xc2\xa74G\xc2\xa7r\xc2\xa76a\xc2\xa7r\xc2\xa7ey\xc2\xa7r\xc2\xa72B\xc2\xa7r\xc2"
        b"\xa71o\xc2\xa7r\xc2\xa79w\xc2\xa7r\xc2\xa7ds\xc2\xa7r\xc2\xa74e\xc2\xa7r\xc2\xa76r;422;;1;69;376707197539105"
        b"3022;;Default;1;19132;-1;"
    )
    parsed = BedrockServerStatus.parse_response(data, 1)
    assert isinstance(parsed, BedrockStatusResponse)


@pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
def test_latency_is_real_number():
    """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

    def mocked_read_status():
        time.sleep(0.001)
        return mock.DEFAULT

    pinger = BedrockServerStatus(Address("localhost", 25565))
    with (
        mock.patch.object(pinger, "_read_status") as mocked_read,
        mock.patch.object(pinger, "parse_response") as mocked_parse_response,
    ):
        mocked_read.side_effect = mocked_read_status

        pinger.read_status()

        # we slept 1ms, so this should be always ~1.
        assert mocked_parse_response.call_args[0][1] >= 1


@pytest.mark.asyncio
@pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
async def test_async_latency_is_real_number():
    """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

    def mocked_read_status():
        time.sleep(0.001)
        return mock.DEFAULT

    pinger = BedrockServerStatus(Address("localhost", 25565))
    with (
        mock.patch.object(pinger, "_read_status_async") as mocked_read,
        mock.patch.object(pinger, "parse_response") as mocked_parse_response,
    ):
        mocked_read.side_effect = mocked_read_status

        await pinger.read_status_async()

        # we slept 1ms, so this should be always ~1.
        assert mocked_parse_response.call_args[0][1] >= 1



from mcstatus import JavaServer

server = JavaServer.lookup("play.hypixel.net")
query = server.query()

if query.players.list:
    print("Players online:", ", ".join(query.players.list))
else:
    status = server.status()

    if not status.players.sample:
        print("Cant find players list, no one online or the server disabled this.")
    else:
        print("Players online:", ", ".join([player.name for player in status.players.sample]))



"""Configuration file for the Sphinx documentation builder.

This file does only contain a selection of the most common options. For a
full list see the documentation:
http://www.sphinx-doc.org/en/master/config
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from __future__ import annotations

import os
import sys

from packaging.version import Version, parse as parse_version
from importlib.metadata import version as importlib_version


sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------


def _get_version() -> Version:
    return parse_version(importlib_version("mcstatus"))


project = "mcstatus"
copyright = "mcstatus, py-mine"
author = "Dinnerbone"

parsed_version = _get_version()

# The short X.Y version
version = parsed_version.base_version
# The full version, including alpha/beta/rc tags
release = str(parsed_version)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    # Used to reference for third party projects:
    "sphinx.ext.intersphinx",
    # Used to include .md files:
    "m2r2",
]

autoclass_content = "both"
autodoc_member_order = "bysource"

autodoc_default_flags = {
    "members": "",
    "undoc-members": "code,error_template",
    "exclude-members": "__dict__,__weakref__",
}

# Automatically generate section labels:
autosectionlabel_prefix_document = True

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:

source_suffix = [".rst", ".md"]

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

add_module_names = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "navigation_with_keys": True,
}

# -- Extension configuration -------------------------------------------------

# Third-party projects documentation references:
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "dns": ("https://dnspython.readthedocs.io/en/stable/", None),
}


# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# Mocks
def mock_autodoc() -> None:
    """Mock autodoc to not add ``Bases: object`` to the classes, that do not have super classes.

    See also https://stackoverflow.com/a/75041544/20952782.
    """
    from sphinx.ext import autodoc

    class MockedClassDocumenter(autodoc.ClassDocumenter):
        def add_line(self, line: str, source: str, *lineno: int) -> None:
            if line == "   Bases: :py:class:`object`":
                return
            super().add_line(line, source, *lineno)

    autodoc.ClassDocumenter = MockedClassDocumenter


def delete_doc_for_address_base() -> None:
    """``__new__`` is appended to the docstring of ``AddressBase``.

    And we do not want autogenerated nonsense docstring there.
    """
    from mcstatus.address import _AddressBase

    del _AddressBase.__new__.__doc__


mock_autodoc()
delete_doc_for_address_base()



from pytest import fixture

from mcstatus.motd import Motd
from mcstatus.responses import LegacyStatusPlayers, LegacyStatusResponse, LegacyStatusVersion
from tests.responses import BaseResponseTest


@fixture(scope="module")
def build():
    return LegacyStatusResponse.build(
        [
            "47",
            "1.4.2",
            "A Minecraft Server",
            "0",
            "20",
        ],
        123.0,
    )


@BaseResponseTest.construct
class TestLegacyStatusResponse(BaseResponseTest):
    EXPECTED_VALUES = [
        ("motd", Motd.parse("A Minecraft Server")),
        ("latency", 123.0),
    ]
    EXPECTED_TYPES = [
        ("players", LegacyStatusPlayers),
        ("version", LegacyStatusVersion),
    ]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build

    def test_as_dict(self, build: LegacyStatusResponse):
        assert build.as_dict() == {
            "latency": 123.0,
            "motd": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.4.2", "protocol": 47},
        }


@BaseResponseTest.construct
class TestLegacyStatusPlayers(BaseResponseTest):
    EXPECTED_VALUES = [("online", 0), ("max", 20)]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build.players


@BaseResponseTest.construct
class TestLegacyStatusVersion(BaseResponseTest):
    EXPECTED_VALUES = [("name", "1.4.2"), ("protocol", 47)]

    @fixture(scope="class")
    def build(self, build):  # pyright: ignore[reportIncompatibleMethodOverride]
        return build.version



from __future__ import annotations

import asyncio
from collections.abc import Iterable
from unittest.mock import call, patch
from typing import SupportsIndex, TypeAlias

import pytest
import pytest_asyncio

from mcstatus.protocol.connection import BaseAsyncReadSyncWriteConnection, Connection
from mcstatus.address import Address
from mcstatus.server import BedrockServer, JavaServer, LegacyServer

BytesConvertable: TypeAlias = "SupportsIndex | Iterable[SupportsIndex]"


class AsyncConnection(BaseAsyncReadSyncWriteConnection):
    def __init__(self) -> None:
        self.sent = bytearray()
        self.received = bytearray()

    async def read(self, length: int) -> bytearray:
        """Return :attr:`.received` up to length bytes, then cut received up to that point."""
        if len(self.received) < length:
            raise IOError(f"Not enough data to read! {len(self.received)} < {length}")

        result = self.received[:length]
        self.received = self.received[length:]
        return result

    def write(self, data: Connection | str | bytearray | bytes) -> None:
        """Extend :attr:`.sent` from ``data``."""
        if isinstance(data, Connection):
            data = data.flush()
        if isinstance(data, str):
            data = bytearray(data, "utf-8")
        self.sent.extend(data)

    def receive(self, data: BytesConvertable | bytearray) -> None:
        """Extend :attr:`.received` with ``data``."""
        if not isinstance(data, bytearray):
            data = bytearray(data)
        self.received.extend(data)

    def remaining(self) -> int:
        """Return length of :attr:`.received`."""
        return len(self.received)

    def flush(self) -> bytearray:
        """Return :attr:`.sent`, also clears :attr:`.sent`."""
        result, self.sent = self.sent, bytearray()
        return result


class MockProtocolFactory(asyncio.Protocol):
    transport: asyncio.Transport

    def __init__(self, data_expected_to_receive, data_to_respond_with):
        self.data_expected_to_receive = data_expected_to_receive
        self.data_to_respond_with = data_to_respond_with

    def connection_made(self, transport: asyncio.Transport):  # pyright: ignore[reportIncompatibleMethodOverride]
        print("connection_made")
        self.transport = transport

    def connection_lost(self, exc):
        print("connection_lost")
        self.transport.close()

    def data_received(self, data):
        assert self.data_expected_to_receive in data
        self.transport.write(self.data_to_respond_with)

    def eof_received(self):
        print("eof_received")

    def pause_writing(self):
        print("pause_writing")

    def resume_writing(self):
        print("resume_writing")


@pytest_asyncio.fixture()
async def create_mock_packet_server():
    event_loop = asyncio.get_running_loop()
    servers = []

    async def create_server(port, data_expected_to_receive, data_to_respond_with):
        server = await event_loop.create_server(
            lambda: MockProtocolFactory(data_expected_to_receive, data_to_respond_with),
            host="localhost",
            port=port,
        )
        servers.append(server)
        return server

    yield create_server

    for server in servers:
        server.close()
        await server.wait_closed()


class TestBedrockServer:
    def setup_method(self):
        self.server = BedrockServer("localhost")

    def test_default_port(self):
        assert self.server.address.port == 19132

    def test_lookup_constructor(self):
        s = BedrockServer.lookup("example.org")
        assert s.address.host == "example.org"
        assert s.address.port == 19132


class TestAsyncJavaServer:
    @pytest.mark.asyncio
    async def test_async_ping(self, unused_tcp_port, create_mock_packet_server):
        await create_mock_packet_server(
            port=unused_tcp_port,
            data_expected_to_receive=bytearray.fromhex("09010000000001C54246"),
            data_to_respond_with=bytearray.fromhex("0F002F096C6F63616C686F737463DD0109010000000001C54246"),
        )
        minecraft_server = JavaServer("localhost", port=unused_tcp_port)

        latency = await minecraft_server.async_ping(ping_token=29704774, version=47)
        assert latency >= 0

    @pytest.mark.asyncio
    async def test_async_lookup_constructor(self):
        s = await JavaServer.async_lookup("example.org:3333")
        assert s.address.host == "example.org"
        assert s.address.port == 3333


def test_java_server_with_query_port():
    with patch("mcstatus.server.JavaServer._retry_query") as patched_query_func:
        server = JavaServer("localhost", query_port=12345)
        server.query()
        assert server.query_port == 12345
        assert patched_query_func.call_args == call(Address("127.0.0.1", port=12345), tries=3)


@pytest.mark.asyncio
async def test_java_server_with_query_port_async():
    with patch("mcstatus.server.JavaServer._retry_async_query") as patched_query_func:
        server = JavaServer("localhost", query_port=12345)
        await server.async_query()
        assert server.query_port == 12345
        assert patched_query_func.call_args == call(Address("127.0.0.1", port=12345), tries=3)


class TestJavaServer:
    def setup_method(self):
        self.socket = Connection()
        self.server = JavaServer("localhost")

    def test_default_port(self):
        assert self.server.address.port == 25565

    def test_ping(self):
        self.socket.receive(bytearray.fromhex("09010000000001C54246"))

        with patch("mcstatus.server.TCPSocketConnection") as connection:
            connection.return_value.__enter__.return_value = self.socket
            latency = self.server.ping(ping_token=29704774, version=47)

        assert self.socket.flush() == bytearray.fromhex("0F002F096C6F63616C686F737463DD0109010000000001C54246")
        assert self.socket.remaining() == 0, "Data is pending to be read, but should be empty"
        assert latency >= 0

    def test_ping_retry(self):
        # Use a blank mock for the connection, we don't want to actually create any connections
        with patch("mcstatus.server.TCPSocketConnection"), patch("mcstatus.server.ServerPinger") as pinger:
            pinger.side_effect = [RuntimeError, RuntimeError, RuntimeError]
            with pytest.raises(RuntimeError):
                self.server.ping()
            assert pinger.call_count == 3

    def test_status(self):
        self.socket.receive(
            bytearray.fromhex(
                "6D006B7B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A7B2"
                "26D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E38222C2270726F746F"
                "636F6C223A34377D7D"
            )
        )

        with patch("mcstatus.server.TCPSocketConnection") as connection:
            connection.return_value.__enter__.return_value = self.socket
            info = self.server.status(version=47)

        assert self.socket.flush() == bytearray.fromhex("0F002F096C6F63616C686F737463DD010100")
        assert self.socket.remaining() == 0, "Data is pending to be read, but should be empty"
        assert info.raw == {
            "description": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.8", "protocol": 47},
        }
        assert info.latency >= 0

    def test_status_retry(self):
        # Use a blank mock for the connection, we don't want to actually create any connections
        with patch("mcstatus.server.TCPSocketConnection"), patch("mcstatus.server.ServerPinger") as pinger:
            pinger.side_effect = [RuntimeError, RuntimeError, RuntimeError]
            with pytest.raises(RuntimeError):
                self.server.status()
            assert pinger.call_count == 3

    def test_query(self):
        self.socket.receive(bytearray.fromhex("090000000035373033353037373800"))
        self.socket.receive(
            bytearray.fromhex(
                "00000000000000000000000000000000686f73746e616d650041204d696e656372616674205365727665720067616d6574797"
                "06500534d500067616d655f6964004d494e4543524146540076657273696f6e00312e3800706c7567696e7300006d61700077"
                "6f726c64006e756d706c61796572730033006d6178706c617965727300323000686f7374706f727400323535363500686f737"
                "46970003139322e3136382e35362e31000001706c617965725f000044696e6e6572626f6e6500446a696e6e69626f6e650053"
                "746576650000"
            )
        )

        with patch("mcstatus.protocol.connection.Connection.remaining") as mock_remaining:
            mock_remaining.side_effect = [15, 208]

            with (
                patch("mcstatus.server.UDPSocketConnection") as connection,
                patch.object(self.server.address, "resolve_ip") as resolve_ip,
            ):
                connection.return_value.__enter__.return_value = self.socket
                resolve_ip.return_value = "127.0.0.1"
                info = self.server.query()

            conn_bytes = self.socket.flush()
            assert conn_bytes[:3] == bytearray.fromhex("FEFD09")
            assert info.raw == {
                "hostname": "A Minecraft Server",
                "gametype": "SMP",
                "game_id": "MINECRAFT",
                "version": "1.8",
                "plugins": "",
                "map": "world",
                "numplayers": "3",
                "maxplayers": "20",
                "hostport": "25565",
                "hostip": "192.168.56.1",
            }

    def test_query_retry(self):
        # Use a blank mock for the connection, we don't want to actually create any connections
        with patch("mcstatus.server.UDPSocketConnection"), patch("mcstatus.server.ServerQuerier") as querier:
            querier.side_effect = [RuntimeError, RuntimeError, RuntimeError]
            with pytest.raises(RuntimeError), patch.object(self.server.address, "resolve_ip") as resolve_ip:
                resolve_ip.return_value = "127.0.0.1"
                self.server.query()
            assert querier.call_count == 3

    def test_lookup_constructor(self):
        s = JavaServer.lookup("example.org:4444")
        assert s.address.host == "example.org"
        assert s.address.port == 4444


class TestLegacyServer:
    def setup_method(self):
        self.socket = Connection()
        self.server = LegacyServer("localhost")

    def test_default_port(self):
        assert self.server.address.port == 25565

    def test_lookup_constructor(self):
        s = LegacyServer.lookup("example.org:4444")
        assert s.address.host == "example.org"
        assert s.address.port == 4444

    def test_status(self):
        self.socket.receive(
            bytearray.fromhex(
                "ff002300a70031000000340037000000"
                "31002e0034002e003200000041002000"
                "4d0069006e0065006300720061006600"
                "74002000530065007200760065007200"
                "000030000000320030"
            )
        )

        with patch("mcstatus.server.TCPSocketConnection") as connection:
            connection.return_value.__enter__.return_value = self.socket
            info = self.server.status()

        assert self.socket.flush() == bytearray.fromhex("fe01fa")
        assert self.socket.remaining() == 0, "Data is pending to be read, but should be empty"
        assert info.as_dict() == {
            "latency": info.latency,
            "motd": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.4.2", "protocol": 47},
        }
        assert info.latency >= 0


class TestAsyncLegacyServer:
    def setup_method(self):
        self.socket = AsyncConnection()
        self.server = LegacyServer("localhost")

    @pytest.mark.asyncio
    async def test_async_lookup_constructor(self):
        s = await LegacyServer.async_lookup("example.org:3333")
        assert s.address.host == "example.org"
        assert s.address.port == 3333

    @pytest.mark.asyncio
    async def test_async_status(self):
        self.socket.receive(
            bytearray.fromhex(
                "ff002300a70031000000340037000000"
                "31002e0034002e003200000041002000"
                "4d0069006e0065006300720061006600"
                "74002000530065007200760065007200"
                "000030000000320030"
            )
        )

        with patch("mcstatus.server.TCPAsyncSocketConnection") as connection:
            connection.return_value.__aenter__.return_value = self.socket
            info = await self.server.async_status()

        assert self.socket.flush() == bytearray.fromhex("fe01fa")
        assert self.socket.remaining() == 0, "Data is pending to be read, but should be empty"
        assert info.as_dict() == {
            "latency": info.latency,
            "motd": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.4.2", "protocol": 47},
        }
        assert info.latency >= 0



from pytest import raises

from mcstatus.responses import BaseStatusResponse


class TestMCStatusResponse:
    def test_raises_not_implemented_error_on_build(self):
        with raises(NotImplementedError):
            BaseStatusResponse.build({"foo": "bar"})  # type: ignore # method is abstract



import warnings
from typing import Any

import pytest

from mcstatus.utils import deprecated


def test_deprecated_function_produces_warning():
    f = deprecated()(lambda: None)

    with warnings.catch_warnings(record=True) as w:
        f()
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message).lower()


def test_deprecated_class_produces_warning():
    test_cls: Any = type("TestCls", (), {"foo": lambda: None})  # https://github.com/microsoft/pyright/discussions/3438
    dep_cls = deprecated(methods=["foo"])(test_cls)

    with warnings.catch_warnings(record=True) as w:
        dep_cls.foo()
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message).lower()


def test_deprecated_function_return_result():
    f = deprecated()(lambda x: x)

    with warnings.catch_warnings(record=True):
        result = f(10)

    assert result == 10


def test_deprecated_class_return_result():
    test_cls: Any = type("TestCls", (), {"foo": lambda x: x})  # https://github.com/microsoft/pyright/discussions/3438
    dep_cls = deprecated(methods=["foo"])(test_cls)

    with warnings.catch_warnings(record=True):
        result = dep_cls.foo(10)

    assert result == 10


def test_deprecated_function_with_methods():
    with pytest.raises(ValueError):
        deprecated(methods=["__str__"])(lambda: None)


def test_deprecated_class_without_methods():
    test_cls: Any = type("TestCls", (), {"foo": lambda x: x})  # https://github.com/microsoft/pyright/discussions/3438
    with pytest.raises(ValueError):
        deprecated()(test_cls)


@pytest.mark.parametrize(
    "arguments",
    [
        {"replacement": "library.newfunction"},
        {"version": "4.8.1"},
        {"date": "2022-06"},
        {"msg": "Why would you still use this?"},
        {"replacement": "new_function", "version": "0.1.2", "msg": "Function got renamed"},
        {"replacement": "new_function", "date": "2022-11", "msg": "Function got renamed"},
    ],
)
def test_deprecated_arguments(arguments):
    f = deprecated(**arguments)(lambda: None)

    with warnings.catch_warnings(record=True) as w:
        f()
        warning_message = str(w[-1].message)

        for value in arguments.values():
            assert value in warning_message


def test_deprecated_no_call_decorator():
    """This makes sure deprecated decorator can run directly on a function without first being called.

    Usually we use deprecated decorator with arguments like ``@deprecated(msg='hi')``, however this means
    that using it normally would usually require doing ``@deprecated()``, but since all of the arguments
    are optional, we can also support using ``@deprecated``, without calling it first. This tests that
    this behavior is in fact working.
    """
    f = deprecated(lambda x: x)

    with warnings.catch_warnings(record=True) as w:
        result = f(10)
        assert issubclass(w[-1].category, DeprecationWarning)
        assert "deprecated" in str(w[-1].message).lower()

    assert result == 10



import sys
import time
from unittest import mock

import pytest

from mcstatus.address import Address
from mcstatus.pinger import ServerPinger
from mcstatus.protocol.connection import Connection


class TestServerPinger:
    def setup_method(self):
        self.pinger = ServerPinger(
            Connection(),  # type: ignore[arg-type]
            address=Address("localhost", 25565),
            version=44,
        )

    def test_handshake(self):
        self.pinger.handshake()

        assert self.pinger.connection.flush() == bytearray.fromhex("0F002C096C6F63616C686F737463DD01")

    def test_read_status(self):
        self.pinger.connection.receive(
            bytearray.fromhex(
                "7200707B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A7B2"
                "26D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E382D70726531222C22"
                "70726F746F636F6C223A34347D7D"
            )
        )
        status = self.pinger.read_status()

        assert status.raw == {
            "description": "A Minecraft Server",
            "players": {"max": 20, "online": 0},
            "version": {"name": "1.8-pre1", "protocol": 44},
        }
        assert self.pinger.connection.flush() == bytearray.fromhex("0100")

    def test_read_status_invalid_json(self):
        self.pinger.connection.receive(bytearray.fromhex("0300017B"))
        with pytest.raises(IOError):
            self.pinger.read_status()

    def test_read_status_invalid_reply(self):
        self.pinger.connection.receive(
            # no motd, see also #922
            bytearray.fromhex(
                "4F004D7B22706C6179657273223A7B226D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616"
                "D65223A22312E382D70726531222C2270726F746F636F6C223A34347D7D"
            )
        )

        self.pinger.read_status()

    def test_read_status_invalid_status(self):
        self.pinger.connection.receive(bytearray.fromhex("0105"))

        with pytest.raises(IOError):
            self.pinger.read_status()

    def test_test_ping(self):
        self.pinger.connection.receive(bytearray.fromhex("09010000000000DD7D1C"))
        self.pinger.ping_token = 14515484

        assert self.pinger.test_ping() >= 0
        assert self.pinger.connection.flush() == bytearray.fromhex("09010000000000DD7D1C")

    def test_test_ping_invalid(self):
        self.pinger.connection.receive(bytearray.fromhex("011F"))
        self.pinger.ping_token = 14515484

        with pytest.raises(IOError):
            self.pinger.test_ping()

    def test_test_ping_wrong_token(self):
        self.pinger.connection.receive(bytearray.fromhex("09010000000000DD7D1C"))
        self.pinger.ping_token = 12345

        with pytest.raises(IOError):
            self.pinger.test_ping()

    @pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
    def test_latency_is_real_number(self):
        """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

        def mocked_read_buffer():
            time.sleep(0.001)
            return mock.DEFAULT

        with mock.patch.object(Connection, "read_buffer") as mocked:
            mocked.side_effect = mocked_read_buffer
            mocked.return_value.read_varint.return_value = 0
            mocked.return_value.read_utf.return_value = """
            {
                "description": "A Minecraft Server",
                "players": {"max": 20, "online": 0},
                "version": {"name": "1.8-pre1", "protocol": 44}
            }
            """
            pinger = ServerPinger(
                Connection(),  # type: ignore[arg-type]
                address=Address("localhost", 25565),
                version=44,
            )

            pinger.connection.receive(
                bytearray.fromhex(
                    "7200707B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A"
                    "7B226D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E382D70726531"
                    "222C2270726F746F636F6C223A34347D7D"
                )
            )
            # we slept 1ms, so this should be always ~1.
            assert pinger.read_status().latency >= 1

    @pytest.mark.flaky(reruns=5, condition=sys.platform.startswith("win32"))
    def test_test_ping_is_in_milliseconds(self):
        """``time.perf_counter`` returns fractional seconds, we must convert it to milliseconds."""

        def mocked_read_buffer():
            time.sleep(0.001)
            return mock.DEFAULT

        with mock.patch.object(Connection, "read_buffer") as mocked:
            mocked.side_effect = mocked_read_buffer
            mocked.return_value.read_varint.return_value = 1
            mocked.return_value.read_long.return_value = 123456789
            pinger = ServerPinger(
                Connection(),  # type: ignore[arg-type]
                address=Address("localhost", 25565),
                version=44,
                ping_token=123456789,
            )

            pinger.connection.receive(
                bytearray.fromhex(
                    "7200707B226465736372697074696F6E223A2241204D696E65637261667420536572766572222C22706C6179657273223A"
                    "7B226D6178223A32302C226F6E6C696E65223A307D2C2276657273696F6E223A7B226E616D65223A22312E382D70726531"
                    "222C2270726F746F636F6C223A34347D7D"
                )
            )
            # we slept 1ms, so this should be always ~1.
            assert pinger.test_ping() >= 1



import pytest

from mcstatus.utils import or_none


@pytest.mark.parametrize(
    "a,b,result",
    [
        (None, None, None),
        (None, "", ""),
        ("", None, ""),
        ("a", "b", "a"),
    ],
)
def test_or_none(a, b, result):
    assert or_none(a, b) == result


def test_or_none_many_arguments():
    assert or_none(*([None] * 100 + ["value"])) == "value"



from __future__ import annotations

import pytest

from mcstatus.motd.components import WebColor


class TestWebColor:
    @pytest.mark.parametrize(
        "hex,rgb",
        [
            ("#bfff00", (191, 255, 0)),
            ("#00ff80", (0, 255, 128)),
            ("#4000ff", (64, 0, 255)),
        ],
    )
    def test_hex_to_rgb_correct(self, hex, rgb):
        assert WebColor.from_hex(hex=hex).rgb == rgb

    @pytest.mark.parametrize(
        "hex,rgb",
        [
            ("#bfff00", (191, 255, 0)),
            ("#00ff80", (0, 255, 128)),
            ("#4000ff", (64, 0, 255)),
        ],
    )
    def test_rgb_to_hex_correct(self, hex, rgb):
        assert WebColor.from_rgb(rgb=rgb).hex == hex

    def test_hex_in_output_has_number_sign(self):
        assert WebColor.from_hex(hex="#bfff00").hex == "#bfff00"
        assert WebColor.from_hex(hex="4000ff").hex == "#4000ff"

    def test_fail_on_incorrect_hex(self):
        with pytest.raises(ValueError):
            WebColor.from_hex(hex="abcd")

    @pytest.mark.parametrize("length", [0, 1, 2, 4, 5, 7, 8, 9, 10])
    def test_fail_on_too_long_or_too_short_hex(self, length: int):
        with pytest.raises(ValueError):
            WebColor.from_hex(hex="a" * length)

    def test_fail_on_incorrect_rgb(self):
        with pytest.raises(ValueError):
            WebColor.from_rgb(rgb=(-23, 699, 1000))

    def test_3_symbols_hex(self):
        assert WebColor.from_hex("a1b").hex == "#aa11bb"



import asyncio

from mcstatus import JavaServer


async def ping_server(ip: str) -> None:
    try:
        status = await (await JavaServer.async_lookup(ip)).async_status()
    except Exception:
        return

    print(f"{ip} - {status.latency}ms")  # handle somehow responses here


async def ping_ips(ips: list[str]) -> None:
    to_process: list[str] = []

    for ip in ips:
        if len(to_process) <= 10:  # 10 means here how many servers will be pinged at once
            to_process.append(ip)
            continue

        await asyncio.wait({asyncio.create_task(ping_server(ip_to_ping)) for ip_to_ping in to_process})
        to_process = []


def main() -> None:
    ips = ["hypixel.net", "play.hivemc.com", "play.cubecraft.net", ...]  # insert here your ips!
    asyncio.run(ping_ips(ips))


if __name__ == "__main__":
    main()



from __future__ import annotations

from contextlib import ExitStack
from unittest import mock

import pytest

from mcstatus.motd import Motd
from mcstatus.motd.components import Formatting, MinecraftColor, TranslationTag, WebColor
from mcstatus.motd.simplifies import (
    get_double_colors,
    get_double_items,
    get_empty_text,
    get_end_non_text,
    get_formatting_before_color,
    get_unused_elements,
)


class TestMotdSimplifies:
    def test_get_unused_elements_call_every_simplifier(self):
        with ExitStack() as stack:
            mocked = []
            for simplifier in [
                get_double_items.__name__,
                get_double_colors.__name__,
                get_formatting_before_color.__name__,
                get_empty_text.__name__,
                get_end_non_text.__name__,
            ]:
                mocked.append(stack.enter_context(mock.patch("mcstatus.motd.simplifies." + simplifier)))

            get_unused_elements([])

            for simplifier in mocked:
                simplifier.assert_called()

    def test_simplify_returns_new_instance(self):
        parsed = ["", Formatting.RESET]
        obj = Motd(parsed.copy(), raw="")
        assert obj.simplify().parsed == []
        assert obj.parsed == parsed

    def test_simplifies_work(self):
        get_unused_elements(["a", "b", "c"])

    def test_simplify_runs_few_times(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#discussion_r1051658497`_."""
        obj = Motd([Formatting.BOLD, "", Formatting.RESET, "", MinecraftColor.RED, ""], raw="")
        assert obj.simplify() == Motd([], raw="")

    @pytest.mark.parametrize("first", (MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    @pytest.mark.parametrize("second", (MinecraftColor.BLUE, WebColor.from_hex(hex="#dd0220")))
    def test_get_double_colors(self, first, second):
        assert get_double_colors([first, second]) == {0}

    @pytest.mark.parametrize("first", (MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    @pytest.mark.parametrize("second", (MinecraftColor.BLUE, WebColor.from_hex(hex="#dd0220")))
    @pytest.mark.parametrize("third", (MinecraftColor.BLUE, WebColor.from_hex(hex="dd0220")))
    def test_get_double_colors_with_three_items(self, first, second, third):
        assert get_double_colors([first, second, third]) == {0, 1}

    @pytest.mark.parametrize("first", (MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    @pytest.mark.parametrize("second", (MinecraftColor.BLUE, WebColor.from_hex(hex="#dd0220")))
    def test_get_double_colors_with_no_double_colors(self, first, second):
        assert get_double_colors([first, "", second]) == set()

    @pytest.mark.parametrize("last_item", (MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    def test_get_formatting_before_color(self, last_item):
        assert get_formatting_before_color([Formatting.BOLD, last_item]) == {0}

    @pytest.mark.parametrize("first_item", (Formatting.RESET, MinecraftColor.RED, WebColor.from_hex(hex="#ff0000"), "abc"))
    def test_get_formatting_before_color_without_formatting_before_color(self, first_item):
        assert get_formatting_before_color([first_item, "abc", MinecraftColor.WHITE]) == set()

    def test_skip_get_formatting_before_color(self):
        assert get_formatting_before_color(["abc", Formatting.BOLD, "def", Formatting.RESET, "ghi"]) == set()

    @pytest.mark.parametrize("last_item", (MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    def test_get_formatting_before_color_if_space_between(self, last_item):
        assert get_formatting_before_color([Formatting.BOLD, " ", last_item]) == {0}

    def test_get_empty_text_removes_empty_string(self):
        assert get_empty_text([Formatting.BOLD, "", Formatting.RESET, "", MinecraftColor.RED, ""]) == {1, 3, 5}

    def test_two_formattings_before_minecraft_color(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#discussion_r1048476090`_."""
        assert get_formatting_before_color([Formatting.BOLD, Formatting.ITALIC, MinecraftColor.RED]) == {0, 1}

    def test_two_formattings_one_by_one(self):
        obj = Motd([Formatting.BOLD, Formatting.ITALIC], raw="")
        assert obj.simplify().parsed == []

    @pytest.mark.parametrize("item", (Formatting.RESET, MinecraftColor.RED, WebColor.from_hex(hex="#ff1234")))
    def test_dont_remove_empty_text(self, item):
        assert get_empty_text([item]) == set()

    @pytest.mark.parametrize("last_item", (Formatting.RESET, MinecraftColor.RED, WebColor.from_hex(hex="#ff0000")))
    def test_non_text_in_the_end(self, last_item):
        assert get_end_non_text(["abc", Formatting.BOLD, "def", Formatting.RESET, "ghi", last_item]) == {5}

    def test_translation_tag_in_the_end(self):
        assert get_end_non_text(["abc", Formatting.BOLD, "def", Formatting.RESET, "ghi", TranslationTag("key")]) == set()

    def test_no_conflict_on_poping_items(self):
        """See `https://github.com/py-mine/mcstatus/pull/335#discussion_r1045303652`_."""
        obj = Motd(["0", "1"], raw="")
        call_count = 0

        def remove_first_element(*_, **__):
            nonlocal call_count
            call_count += 1
            if call_count in (1, 2):
                return {0}
            return set()

        with ExitStack() as stack:
            for simplifier in [
                get_double_items.__name__,
                get_double_colors.__name__,
                get_formatting_before_color.__name__,
                get_empty_text.__name__,
                get_end_non_text.__name__,
            ]:
                stack.enter_context(mock.patch("mcstatus.motd.simplifies." + simplifier, remove_first_element))
            assert obj.simplify().parsed == ["1"]

    def test_simplify_function_provides_the_same_raw(self):
        obj = object()
        assert Motd([], raw=obj).simplify().raw is obj  # type: ignore # Invalid argument type

    def test_simplify_do_not_remove_string_contains_only_spaces(self):
        """Those can be used as delimiters."""
        assert Motd([" " * 20], raw="").simplify().parsed == [" " * 20]

    def test_simplify_meaningless_resets_and_colors(self):
        assert Motd.parse("&a1&a2&a3").simplify().parsed == [MinecraftColor.GREEN, "123"]

    def test_remove_formatting_reset_if_there_was_no_color_or_formatting(self):
        motd = Motd.parse({"text": "123", "extra": [{"text": "123"}]})
        assert motd.parsed == ["123", Formatting.RESET, "123", Formatting.RESET]
        assert motd.simplify().parsed == ["123123"]

    def test_squash_nearby_strings(self):
        assert Motd(["123", "123", "123"], raw="").simplify().parsed == ["123123123"]



from __future__ import annotations

import typing
from collections.abc import Callable

import pytest

from mcstatus.motd import Motd
from mcstatus.motd.transformers import AnsiTransformer, HtmlTransformer, MinecraftTransformer, PlainTransformer

if typing.TYPE_CHECKING:
    from mcstatus.responses import RawJavaResponseMotd


class TestMotdPlain:
    @pytest.fixture(scope="class", params=["attribute", "function"])
    def result(self, request) -> Callable[[str | RawJavaResponseMotd], str]:
        if request.param == "attribute":
            return lambda text: Motd.parse(text).to_plain()
        else:
            return lambda text: PlainTransformer().transform(Motd.parse(text).parsed)

    def test_plain_text(self, result):
        assert result("plain") == "plain"

    def test_removes_colors(self, result):
        assert result("&1&ltext") == "text"

    def test_skip_web_colors(self, result):
        assert result({"extra": [{"color": "#4000ff", "text": "colored text"}], "text": ""}) == "colored text"

    def test_skip_minecraft_colors(self, result):
        assert result({"extra": [{"color": "red", "text": "colored text"}], "text": ""}) == "colored text"


class TestMotdMinecraft:
    @pytest.fixture(scope="class", params=["attribute", "function"])
    def result(self, request) -> Callable[[str | RawJavaResponseMotd], str]:
        if request.param == "attribute":
            return lambda text: Motd.parse(text).to_minecraft()
        else:
            return lambda text: MinecraftTransformer().transform(Motd.parse(text).parsed)

    @pytest.mark.parametrize("motd", ["&1&2&3", "§123§5bc", "§1§2§3"])
    def test_return_the_same(self, motd: str, result):
        assert result(motd) == motd.replace("&", "§")

    def test_skip_web_colors(self, result):
        assert result({"extra": [{"color": "#4000ff", "text": "colored text"}], "text": ""}) == "§rcolored text§r"


class TestMotdHTML:
    @pytest.fixture(scope="class", params=["attribute", "class"])
    def result(self, request) -> Callable[[str, bool], str]:
        if request.param == "attribute":
            return lambda text, bedrock: Motd.parse(text, bedrock=bedrock).to_html()
        else:
            return lambda text, bedrock: HtmlTransformer(bedrock=bedrock).transform(Motd.parse(text, bedrock=bedrock).parsed)

    @pytest.mark.parametrize("bedrock", (True, False))
    def test_correct_output(self, result: Callable[["str | dict", bool], str], source, bedrock: bool):
        assert result(source, bedrock) == (
            "<p>top"
            "1<span style='color:rgb(179, 238, 255)'>2</span>"
            "<span style='color:rgb(0, 0, 0);text-shadow:0 0 1px rgb(0, 0, 0)'><span class=obfuscated>3</span>"
            "</span>"
            "<span style='color:rgb(0, 0, 170);text-shadow:0 0 1px rgb(0, 0, 42)'><b><s>4</span></b></s>"
            "<span style='color:rgb(0, 170, 0);text-shadow:0 0 1px rgb(0, 42, 0)'><i>5</span></i>"
            "<span style='color:rgb(0, 170, 170);text-shadow:0 0 1px rgb(0, 42, 42)'><u>6</span></u>"
            "<span style='color:rgb(0, 170, 170);text-shadow:0 0 1px rgb(0, 42, 42)'>7</span>"
            "<span style='color:rgb(170, 0, 0);text-shadow:0 0 1px rgb(42, 0, 0)'>8</span>"
            "<span style='color:rgb(170, 0, 170);text-shadow:0 0 1px rgb(42, 0, 42)'>9</span>"
            "<span style='color:rgb(255, 170, 0);text-shadow:0 0 1px rgb"
            + str((64, 42, 0) if bedrock else (42, 42, 0))
            + "'>10</span>"
            "<span style='color:rgb(170, 170, 170);text-shadow:0 0 1px rgb(42, 42, 42)'>11</span>"
            "<span style='color:rgb(85, 85, 85);text-shadow:0 0 1px rgb(21, 21, 21)'>12</span>"
            "<span style='color:rgb(85, 85, 255);text-shadow:0 0 1px rgb(21, 21, 63)'>13</span>"
            "<span style='color:rgb(85, 255, 85);text-shadow:0 0 1px rgb(21, 63, 21)'>14</span>"
            "<span style='color:rgb(85, 255, 255);text-shadow:0 0 1px rgb(21, 63, 63)'>15</span>"
            "<span style='color:rgb(255, 85, 85);text-shadow:0 0 1px rgb(63, 21, 21)'>16</span>"
            "<span style='color:rgb(255, 85, 255);text-shadow:0 0 1px rgb(63, 21, 63)'>17</span>"
            "<span style='color:rgb(255, 255, 85);text-shadow:0 0 1px rgb(63, 63, 21)'>18</span>"
            "<span style='color:rgb(255, 255, 255);text-shadow:0 0 1px rgb(63, 63, 63)'>19</span>"
            "<span style='color:rgb(221, 214, 5);text-shadow:0 0 1px rgb(55, 53, 1)'>20</span>"
            "21</p>"
        )


class TestMotdAnsi:
    @pytest.fixture(scope="class", params=[True, False])
    def bedrock(self, request):
        return request.param

    @pytest.fixture(scope="class")
    def expected_result(self):
        return (
            "\033[0mtop\033[0m"
            "1\033[0m"
            "\033[38;2;179;238;255m2\033[0m\033[0m"
            "\033[38;2;0;0;0m\033[5m3\033[0m\033[0m"
            "\033[38;2;0;0;170m\033[1m\033[9m4\033[0m\033[0m"
            "\033[38;2;0;170;0m\033[3m5\033[0m\033[0m"
            "\033[38;2;0;170;170m\033[4m6\033[0m\033[0m"
            "\033[38;2;0;170;170m7\033[0m\033[0m"
            "\033[38;2;170;0;0m8\033[0m\033[0m"
            "\033[38;2;170;0;170m9\033[0m\033[0m"
            "\033[38;2;255;170;0m10\033[0m\033[0m"
            "\033[38;2;170;170;170m11\033[0m\033[0m"
            "\033[38;2;85;85;85m12\033[0m\033[0m"
            "\033[38;2;85;85;255m13\033[0m\033[0m"
            "\033[38;2;85;255;85m14\033[0m\033[0m"
            "\033[38;2;85;255;255m15\033[0m\033[0m"
            "\033[38;2;255;85;85m16\033[0m\033[0m"
            "\033[38;2;255;85;255m17\033[0m\033[0m"
            "\033[38;2;255;255;85m18\033[0m\033[0m"
            "\033[38;2;255;255;255m19\033[0m\033[0m"
            "\033[38;2;221;214;5m20\033[0m\033[0m"
            "21\033[0m"
            "\033[0m\033[0m"
        )

    @pytest.fixture(scope="class", params=["attribute", "class"])
    def result(self, request) -> Callable[[str, bool], str]:
        if request.param == "attribute":
            return lambda text, bedrock: Motd.parse(text, bedrock=bedrock).to_ansi()
        else:
            return lambda text, bedrock: AnsiTransformer().transform(Motd.parse(text, bedrock=bedrock).parsed)

    def test_correct_output(self, result: Callable[[str | dict, bool], str], source, bedrock, expected_result):
        assert result(source, bedrock) == expected_result



import io
import socket

from mcstatus import JavaServer, LegacyServer, BedrockServer
from mcstatus.responses import JavaStatusResponse, LegacyStatusResponse, BedrockStatusResponse, RawJavaResponse
from mcstatus.querier import QueryResponse

import os
import json
import contextlib
from unittest import mock
from unittest.mock import patch
import pytest

from mcstatus.__main__ import main as main_under_test, PING_PACKET_FAIL_WARNING, QUERY_FAIL_WARNING

JAVA_RAW_RESPONSE: RawJavaResponse = {
    "players": {"max": 20, "online": 0},
    "version": {"name": "1.8-pre1", "protocol": 44},
    "description": "A Minecraft Server",
    "enforcesSecureChat": True,
    "favicon": "data:image/png;base64,foo",
}

QUERY_RAW_RESPONSE = [
    {
        "hostname": "A Minecraft Server",
        "gametype": "GAME TYPE",
        "game_id": "GAME ID",
        "version": "1.8",
        "plugins": "",
        "map": "world",
        "numplayers": "3",
        "maxplayers": "20",
        "hostport": "9999",
        "hostip": "192.168.56.1",
    },
    ["Dinnerbone", "Djinnibone", "Steve"],
]

LEGACY_RAW_RESPONSE = [
    "47",
    "1.4.2",
    "A Minecraft Server",
    "0",
    "20",
]

BEDROCK_RAW_RESPONSE = [
    "MCPE",
    "§r§4G§r§6a§r§ey§r§2B§r§1o§r§9w§r§ds§r§4e§r§6r",
    "422",
    "1.18.100500",
    "1",
    "69",
    "3767071975391053022",
    "map name here",
    "Default",
    "1",
    "19132",
    "-1",
    "3",
]

# NOTE: if updating this, be sure to change other occurrences of this help text!
# to update, use: `COLUMNS=100000 poetry run mcstatus --help`
EXPECTED_HELP_OUTPUT = """
usage: mcstatus [-h] [--bedrock | --legacy] address {ping,status,query,json} ...

mcstatus provides an easy way to query Minecraft servers for any information they can expose. It provides three modes of access: query, status, ping and json.

positional arguments:
  address               The address of the server.

options:
  -h, --help            show this help message and exit
  --bedrock             Specifies that 'address' is a Bedrock server (default: Java).
  --legacy              Specifies that 'address' is a pre-1.7 Java server (default: 1.7+).

commands:
  Command to run, defaults to 'status'.

  {ping,status,query,json}
    ping                Ping server for latency.
    status              Prints server status.
    query               Prints detailed server information. Must be enabled in servers' server.properties file.
    json                Prints server status and query in json.
"""  # noqa: E501 (line length)


@contextlib.contextmanager
def patch_stdout_stderr():
    outpatch = patch("sys.stdout", new=io.StringIO())
    errpatch = patch("sys.stderr", new=io.StringIO())
    with outpatch as out, errpatch as err:
        yield out, err


@pytest.fixture
def mock_network_requests():
    with (
        patch("mcstatus.server.JavaServer.lookup", return_value=JavaServer("example.com", port=25565)),
        patch("mcstatus.server.JavaServer.ping", return_value=0),
        patch("mcstatus.server.JavaServer.status", return_value=JavaStatusResponse.build(JAVA_RAW_RESPONSE)),
        patch("mcstatus.server.JavaServer.query", return_value=QueryResponse.build(*QUERY_RAW_RESPONSE)),
        patch("mcstatus.server.LegacyServer.lookup", return_value=LegacyServer("example.com", port=25565)),
        patch(
            "mcstatus.server.LegacyServer.status", return_value=LegacyStatusResponse.build(LEGACY_RAW_RESPONSE, latency=123)
        ),
        patch("mcstatus.server.BedrockServer.lookup", return_value=BedrockServer("example.com", port=25565)),
        patch(
            "mcstatus.server.BedrockServer.status",
            return_value=(BedrockStatusResponse.build(BEDROCK_RAW_RESPONSE, latency=123)),
        ),
    ):
        yield


def normalise_help_output(s: str) -> str:
    """
    Normalises the output of `mcstatus --help`, to work around
    some discrepancies between Python versions while still retaining
    meaningful information for comparison.
    """

    elided = "[...]:"

    s = s.strip()

    # drop lines which end in ":". these argparse section headings vary between python versions.
    # it is just a small style change, so it doesn't matter so much to do `sys.version_info` check
    return "\n".join(ln if not ln.endswith(":") else elided for ln in s.splitlines())


# NOTE: for premature exits in argparse, we must catch SystemExit.
# for ordinary exits in the CLI code, we can simply inspect the return value.


def test_no_args():
    with patch_stdout_stderr() as (out, err), pytest.raises(SystemExit) as exn:
        main_under_test([])

    assert out.getvalue() == ""
    assert "usage: " in err.getvalue()
    assert exn.value.code != 0


def test_help():
    with patch_stdout_stderr() as (out, err), pytest.raises(SystemExit) as exn:
        main_under_test(["--help"])

    assert "usage: " in out.getvalue()
    assert err.getvalue() == ""
    assert exn.value.code == 0


@mock.patch.dict(os.environ, {"COLUMNS": "100000"})  # prevent line-wrapping in --help output
def test_help_matches_recorded_output():
    with patch_stdout_stderr() as (out, err), pytest.raises(SystemExit):
        main_under_test(["--help"])

    assert normalise_help_output(out.getvalue()) == normalise_help_output(EXPECTED_HELP_OUTPUT)
    assert err.getvalue() == ""


def test_one_argument_is_status(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com"]) == 0

    assert (
        "version: Java 1.8-pre1 (protocol 44)\n"
        "motd: \x1b[0mA Minecraft Server\x1b[0m\n"
        "players: 0/20 No players online\n"
        "ping: 0.00 ms\n"
    ) == out.getvalue()
    assert err.getvalue() == ""


def test_status(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "status"]) == 0

    assert (
        "version: Java 1.8-pre1 (protocol 44)\n"
        "motd: \x1b[0mA Minecraft Server\x1b[0m\n"
        "players: 0/20 No players online\n"
        "ping: 0.00 ms\n"
    ) == out.getvalue()
    assert err.getvalue() == ""


def test_status_bedrock(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "--bedrock", "status"]) == 0

    assert (
        "version: Bedrock 1.18.100500 (protocol 422)\n"
        "motd: \x1b[0m\x1b[0m\x1b[0m\x1b[38;2;170;0;0mG\x1b[0m\x1b[0m\x1b[38;2;255;170;0ma\x1b[0m\x1b[0m\x1b[38;2;255;255;85m"
        "y\x1b[0m\x1b[0m\x1b[38;2;0;170;0mB\x1b[0m\x1b[0m\x1b[38;2;0;0;170mo\x1b[0m\x1b[0m\x1b[38;2;85;85;255mw\x1b[0m\x1b[0m"
        "\x1b[38;2;255;85;255ms\x1b[0m\x1b[0m\x1b[38;2;170;0;0me\x1b[0m\x1b[0m\x1b[38;2;255;170;0mr\x1b[0m\n"
        "players: 1/69\n"
        "ping: 123.00 ms\n"
    ) == out.getvalue()
    assert err.getvalue() == ""


def test_status_legacy(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "--legacy", "status"]) == 0

    assert (
        "version: Java (pre-1.7) 1.4.2 (protocol 47)\nmotd: \x1b[0mA Minecraft Server\x1b[0m\nplayers: 0/20\nping: 123.00 ms\n"
    ) == out.getvalue()
    assert err.getvalue() == ""


def test_status_offline(mock_network_requests):
    with patch_stdout_stderr() as (out, err), patch("mcstatus.server.JavaServer.status", side_effect=TimeoutError):
        assert main_under_test(["example.com", "status"]) == 1

    assert out.getvalue() == ""
    assert err.getvalue() == "Error: TimeoutError()\n"


def test_query(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "query"]) == 0

    assert (
        "host: 192.168.56.1:9999\n"
        "software: Java 1.8 vanilla\n"
        "motd: \x1b[0mA Minecraft Server\x1b[0m\n"
        "plugins: []\n"
        "players: 3/20 ['Dinnerbone', 'Djinnibone', 'Steve']\n"
    ) == out.getvalue()
    assert err.getvalue() == ""


def test_query_offline(mock_network_requests):
    with patch_stdout_stderr() as (out, err), patch("mcstatus.server.JavaServer.query", side_effect=socket.timeout):
        assert main_under_test(["example.com", "query"]) != 0

    assert out.getvalue() == ""
    assert err.getvalue() == QUERY_FAIL_WARNING + "\n"


def test_json(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "json"]) == 0

    data = json.loads(out.getvalue())
    assert data == {
        "online": True,
        "kind": "Java",
        "status": {
            "players": {"online": 0, "max": 20, "sample": None},
            "version": {"name": "1.8-pre1", "protocol": 44},
            "motd": "A Minecraft Server",
            "latency": 0,
            "raw": {
                "players": {"max": 20, "online": 0},
                "version": {"name": "1.8-pre1", "protocol": 44},
                "description": "A Minecraft Server",
                "enforcesSecureChat": True,
                "favicon": "data:image/png;base64,foo",
            },
            "enforces_secure_chat": True,
            "icon": "data:image/png;base64,foo",
            "forge_data": None,
        },
        "query": {
            "ip": "192.168.56.1",
            "port": "9999",
            "map": "world",
            "plugins": [],
            "raw": {
                "hostname": "A Minecraft Server",
                "gametype": "GAME TYPE",
                "game_id": "GAME ID",
                "version": "1.8",
                "plugins": "",
                "map": "world",
                "numplayers": "3",
                "maxplayers": "20",
                "hostport": "9999",
                "hostip": "192.168.56.1",
            },
        },
    }
    assert err.getvalue() == ""


def test_ping(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "ping"]) == 0

    assert float(out.getvalue()) == 0
    assert err.getvalue() == ""


def test_ping_bedrock(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "--bedrock", "ping"]) == 0

    assert float(out.getvalue()) == 123
    assert err.getvalue() == ""


def test_ping_legacy(mock_network_requests):
    with patch_stdout_stderr() as (out, err):
        assert main_under_test(["example.com", "--legacy", "ping"]) == 0

    assert float(out.getvalue()) == 123
    assert err.getvalue() == ""


def test_ping_server_doesnt_support(mock_network_requests):
    with patch_stdout_stderr() as (out, err), patch("mcstatus.server.JavaServer.ping", side_effect=TimeoutError("timeout")):
        assert main_under_test(["example.com", "ping"]) == 0

    assert float(out.getvalue()) == 0
    assert err.getvalue() == PING_PACKET_FAIL_WARNING.format(address="example.com:25565", ping_exc="timeout") + "\n"



from __future__ import annotations

import dns.resolver
import sys
import json
import argparse
import socket
import dataclasses
from typing import TypeAlias

from mcstatus import JavaServer, LegacyServer, BedrockServer
from mcstatus.responses import JavaStatusResponse
from mcstatus.motd import Motd

SupportedServers: TypeAlias = "JavaServer | LegacyServer | BedrockServer"

PING_PACKET_FAIL_WARNING = (
    "warning: contacting {address} failed with a 'ping' packet but succeeded with a 'status' packet,\n"
    "         this is likely a bug in the server-side implementation.\n"
    '         (note: ping packet failed due to "{ping_exc}")\n'
    "         for more details, see: https://mcstatus.readthedocs.io/en/stable/pages/faq/\n"
)

QUERY_FAIL_WARNING = (
    "The server did not respond to the query protocol."
    "\nPlease ensure that the server has enable-query turned on,"
    " and that the necessary port (same as server-port unless query-port is set) is open in any firewall(s)."
    "\nSee https://minecraft.wiki/w/Query for further information."
)


def _motd(motd: Motd) -> str:
    """Formats MOTD for human-readable output, with leading line break
    if multiline."""
    s = motd.to_ansi()
    return f"\n{s}" if "\n" in s else f" {s}"


def _kind(serv: SupportedServers) -> str:
    if isinstance(serv, JavaServer):
        return "Java"
    elif isinstance(serv, LegacyServer):
        return "Java (pre-1.7)"
    elif isinstance(serv, BedrockServer):
        return "Bedrock"
    else:
        raise ValueError(f"unsupported server for kind: {serv}")


def _ping_with_fallback(server: SupportedServers) -> float:
    # only Java has ping method
    if not isinstance(server, JavaServer):
        return server.status().latency

    # try faster ping packet first, falling back to status with a warning.
    ping_exc = None
    try:
        return server.ping(tries=1)
    except Exception as e:
        ping_exc = e

    latency = server.status().latency

    address = f"{server.address.host}:{server.address.port}"
    print(
        PING_PACKET_FAIL_WARNING.format(address=address, ping_exc=ping_exc),
        file=sys.stderr,
    )

    return latency


def ping_cmd(server: SupportedServers) -> int:
    print(_ping_with_fallback(server))
    return 0


def status_cmd(server: SupportedServers) -> int:
    response = server.status()

    java_res = response if isinstance(response, JavaStatusResponse) else None

    if not java_res:
        player_sample = ""
    elif java_res.players.sample is not None:
        player_sample = str([f"{player.name} ({player.id})" for player in java_res.players.sample])
    else:
        player_sample = "No players online"

    if player_sample:
        player_sample = " " + player_sample

    print(f"version: {_kind(server)} {response.version.name} (protocol {response.version.protocol})")
    print(f"motd:{_motd(response.motd)}")
    print(f"players: {response.players.online}/{response.players.max}{player_sample}")
    print(f"ping: {response.latency:.2f} ms")
    return 0


def json_cmd(server: SupportedServers) -> int:
    data = {"online": False, "kind": _kind(server)}

    status_res = query_res = exn = None
    try:
        status_res = server.status(tries=1)
    except Exception as e:
        exn = exn or e

    try:
        if isinstance(server, JavaServer):
            query_res = server.query(tries=1)
    except Exception as e:
        exn = exn or e

    # construct 'data' dict outside try/except to ensure data processing errors
    # are noticed.
    data["online"] = bool(status_res or query_res)
    if not data["online"]:
        assert exn, "server offline but no exception?"
        data["error"] = str(exn)

    if status_res is not None:
        data["status"] = dataclasses.asdict(status_res)

        # ensure we are overwriting the motd and not making a new dict field
        assert "motd" in data["status"], "motd field missing. has it been renamed?"
        data["status"]["motd"] = status_res.motd.simplify().to_minecraft()

    if query_res is not None:
        # TODO: QueryResponse is not (yet?) a dataclass
        data["query"] = qdata = {}

        qdata["ip"] = query_res.raw["hostip"]
        qdata["port"] = query_res.raw["hostport"]
        qdata["map"] = query_res.map_name
        qdata["plugins"] = query_res.software.plugins
        qdata["raw"] = query_res.raw

    json.dump(data, sys.stdout)
    return 0


def query_cmd(server: SupportedServers) -> int:
    if not isinstance(server, JavaServer):
        print("The 'query' protocol is only supported by Java servers.", file=sys.stderr)
        return 1

    try:
        response = server.query()
    except socket.timeout:
        print(QUERY_FAIL_WARNING, file=sys.stderr)
        return 1

    print(f"host: {response.raw['hostip']}:{response.raw['hostport']}")
    print(f"software: {_kind(server)} {response.software.version} {response.software.brand}")
    print(f"motd:{_motd(response.motd)}")
    print(f"plugins: {response.software.plugins}")
    print(f"players: {response.players.online}/{response.players.max} {response.players.list}")
    return 0


def main(argv: list[str] = sys.argv[1:]) -> int:
    parser = argparse.ArgumentParser(
        "mcstatus",
        description="""
        mcstatus provides an easy way to query Minecraft servers for any information
        they can expose. It provides three modes of access: query, status, ping and json.
        """,
    )

    parser.add_argument("address", help="The address of the server.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--bedrock", help="Specifies that 'address' is a Bedrock server (default: Java).", action="store_true")
    group.add_argument(
        "--legacy", help="Specifies that 'address' is a pre-1.7 Java server (default: 1.7+).", action="store_true"
    )

    subparsers = parser.add_subparsers(title="commands", description="Command to run, defaults to 'status'.")
    parser.set_defaults(func=status_cmd)

    subparsers.add_parser("ping", help="Ping server for latency.").set_defaults(func=ping_cmd)
    subparsers.add_parser("status", help="Prints server status.").set_defaults(func=status_cmd)
    subparsers.add_parser(
        "query", help="Prints detailed server information. Must be enabled in servers' server.properties file."
    ).set_defaults(func=query_cmd)
    subparsers.add_parser(
        "json",
        help="Prints server status and query in json.",
    ).set_defaults(func=json_cmd)

    args = parser.parse_args(argv)
    if args.bedrock:
        lookup = BedrockServer.lookup
    elif args.legacy:
        lookup = LegacyServer.lookup
    else:
        lookup = JavaServer.lookup

    try:
        server = lookup(args.address)
        return args.func(server)
    except (socket.timeout, socket.gaierror, dns.resolver.NoNameservers, ConnectionError, TimeoutError) as e:
        # catch and hide traceback for expected user-facing errors
        print(f"Error: {e!r}", file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



from time import perf_counter

from mcstatus.protocol.connection import BaseAsyncReadSyncWriteConnection, BaseSyncConnection
from mcstatus.responses import LegacyStatusResponse


class _BaseLegacyServerStatus:
    request_status_data = bytes.fromhex(
        # see https://minecraft.wiki/w/Java_Edition_protocol/Server_List_Ping#Client_to_server
        "fe01fa"
    )

    @staticmethod
    def parse_response(data: bytes, latency: float) -> LegacyStatusResponse:
        decoded_data = data.decode("UTF-16BE").split("\0")
        if decoded_data[0] != "§1":
            raise IOError("Recieved invalid kick packet reason")

        return LegacyStatusResponse.build(decoded_data[1:], latency)


class LegacyServerStatus(_BaseLegacyServerStatus):
    def __init__(self, connection: BaseSyncConnection):
        self.connection = connection

    def read_status(self) -> LegacyStatusResponse:
        """Send the status request and read the response."""
        start = perf_counter()
        self.connection.write(self.request_status_data)
        id = self.connection.read(1)
        if id != b"\xff":
            raise IOError("Received invalid packet ID")
        length = self.connection.read_ushort()
        data = self.connection.read(length * 2)
        end = perf_counter()
        return self.parse_response(data, (end - start) * 1000)


class AsyncLegacyServerStatus(_BaseLegacyServerStatus):
    def __init__(self, connection: BaseAsyncReadSyncWriteConnection):
        self.connection = connection

    async def read_status(self) -> LegacyStatusResponse:
        """Send the status request and read the response."""
        start = perf_counter()
        self.connection.write(self.request_status_data)
        id = await self.connection.read(1)
        if id != b"\xff":
            raise IOError("Received invalid packet ID")
        length = await self.connection.read_ushort()
        data = await self.connection.read(length * 2)
        end = perf_counter()
        return self.parse_response(data, (end - start) * 1000)



from mcstatus.motd import Motd
from mcstatus.protocol.connection import Connection
from mcstatus.querier import ServerQuerier


class TestMinecraftQuerier:
    def setup_method(self):
        self.querier = ServerQuerier(Connection())  # type: ignore[arg-type]

    def test_handshake(self):
        self.querier.connection.receive(bytearray.fromhex("090000000035373033353037373800"))
        self.querier.handshake()

        conn_bytes = self.querier.connection.flush()
        assert conn_bytes[:3] == bytearray.fromhex("FEFD09")
        assert self.querier.challenge == 570350778

    def test_query(self):
        self.querier.connection.receive(
            bytearray.fromhex(
                "00000000000000000000000000000000686f73746e616d650041204d696e656372616674205365727665720067616d6574797"
                "06500534d500067616d655f6964004d494e4543524146540076657273696f6e00312e3800706c7567696e7300006d61700077"
                "6f726c64006e756d706c61796572730033006d6178706c617965727300323000686f7374706f727400323535363500686f737"
                "46970003139322e3136382e35362e31000001706c617965725f000044696e6e6572626f6e6500446a696e6e69626f6e650053"
                "746576650000"
            )
        )
        response = self.querier.read_query()
        conn_bytes = self.querier.connection.flush()
        assert conn_bytes[:3] == bytearray.fromhex("FEFD00")
        assert conn_bytes[7:] == bytearray.fromhex("0000000000000000")
        assert response.raw == {
            "hostname": "A Minecraft Server",
            "gametype": "SMP",
            "game_id": "MINECRAFT",
            "version": "1.8",
            "plugins": "",
            "map": "world",
            "numplayers": "3",
            "maxplayers": "20",
            "hostport": "25565",
            "hostip": "192.168.56.1",
        }
        assert response.players.list == ["Dinnerbone", "Djinnibone", "Steve"]

    def test_query_handles_unorderd_map_response(self):
        self.querier.connection.receive(
            bytearray(
                b"\x00\x00\x00\x00\x00GeyserMC\x00\x80\x00hostname\x00Geyser\x00hostip\x001.1.1.1\x00plugins\x00\x00numplayers"
                b"\x001\x00gametype\x00SMP\x00maxplayers\x00100\x00hostport\x0019132\x00version\x00Geyser"
                b" (git-master-0fd903e) 1.18.10\x00map\x00Geyser\x00game_id\x00MINECRAFT\x00\x00\x01player_\x00\x00\x00"
            )
        )
        response = self.querier.read_query()
        self.querier.connection.flush()

        assert response.raw["game_id"] == "MINECRAFT"
        assert response.motd == Motd.parse("Geyser")
        assert response.software.version == "Geyser (git-master-0fd903e) 1.18.10"

    def test_query_handles_unicode_motd_with_nulls(self):
        self.querier.connection.receive(
            bytearray(
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00hostname\x00\x00*K\xd5\x00gametype\x00SMP"
                b"\x00game_id\x00MINECRAFT\x00version\x001.16.5\x00plugins\x00Paper on 1.16.5-R0.1-SNAPSHOT\x00map\x00world"
                b"\x00numplayers\x000\x00maxplayers\x0020\x00hostport\x0025565\x00hostip\x00127.0.1.1\x00\x00\x01player_\x00"
                b"\x00\x00"
            )
        )
        response = self.querier.read_query()
        self.querier.connection.flush()

        assert response.raw["game_id"] == "MINECRAFT"
        assert response.motd == Motd.parse("\x00*KÕ")

    def test_query_handles_unicode_motd_with_2a00_at_the_start(self):
        self.querier.connection.receive(
            bytearray.fromhex(
                "00000000000000000000000000000000686f73746e616d6500006f746865720067616d657479706500534d500067616d655f6964004d"
                "494e4543524146540076657273696f6e00312e31382e3100706c7567696e7300006d617000776f726c64006e756d706c617965727300"
                "30006d6178706c617965727300323000686f7374706f727400323535363500686f73746970003137322e31372e302e32000001706c61"
                "7965725f000000"
            )
        )
        response = self.querier.read_query()
        self.querier.connection.flush()

        assert response.raw["game_id"] == "MINECRAFT"
        assert response.motd == Motd.parse("\x00other")  # "\u2a00other" is actually what is expected,
        # but the query protocol for vanilla has a bug when it comes to unicode handling.
        # The status protocol correctly shows "⨀other".



import pytest

from mcstatus.protocol.connection import Connection
from mcstatus.legacy_status import LegacyServerStatus


def test_invalid_kick_reason():
    with pytest.raises(IOError):
        LegacyServerStatus.parse_response("Invalid Reason".encode("UTF-16BE"), 123.0)


def test_invalid_packet_id():
    socket = Connection()
    socket.receive(bytearray.fromhex("00"))
    server = LegacyServerStatus(socket)
    with pytest.raises(IOError):
        server.read_status()



from __future__ import annotations

import random
import re
import struct
from abc import abstractmethod
from collections.abc import Awaitable
from dataclasses import dataclass, field
from typing import ClassVar, final

from mcstatus.protocol.connection import Connection, UDPAsyncSocketConnection, UDPSocketConnection
from mcstatus.responses import QueryResponse, RawQueryResponse


@dataclass
class _BaseServerQuerier:
    MAGIC_PREFIX: ClassVar = bytearray.fromhex("FEFD")
    PADDING: ClassVar = bytearray.fromhex("00000000")
    PACKET_TYPE_CHALLENGE: ClassVar = 9
    PACKET_TYPE_QUERY: ClassVar = 0

    connection: UDPSocketConnection | UDPAsyncSocketConnection
    challenge: int = field(init=False, default=0)

    @staticmethod
    def _generate_session_id() -> int:
        # minecraft only supports lower 4 bits
        return random.randint(0, 2**31) & 0x0F0F0F0F

    def _create_packet(self) -> Connection:
        packet = Connection()
        packet.write(self.MAGIC_PREFIX)
        packet.write(struct.pack("!B", self.PACKET_TYPE_QUERY))
        packet.write_uint(self._generate_session_id())
        packet.write_int(self.challenge)
        packet.write(self.PADDING)
        return packet

    def _create_handshake_packet(self) -> Connection:
        packet = Connection()
        packet.write(self.MAGIC_PREFIX)
        packet.write(struct.pack("!B", self.PACKET_TYPE_CHALLENGE))
        packet.write_uint(self._generate_session_id())
        return packet

    @abstractmethod
    def _read_packet(self) -> Connection | Awaitable[Connection]:
        raise NotImplementedError

    @abstractmethod
    def handshake(self) -> None | Awaitable[None]:
        raise NotImplementedError

    @abstractmethod
    def read_query(self) -> QueryResponse | Awaitable[QueryResponse]:
        raise NotImplementedError

    def _parse_response(self, response: Connection) -> tuple[RawQueryResponse, list[str]]:
        """Transform the connection object (the result) into dict which is passed to the QueryResponse constructor.

        :return: A tuple with two elements. First is `raw` answer and second is list of players.
        """
        response.read(len("splitnum") + 3)
        data = {}

        while True:
            key = response.read_ascii()
            if key == "hostname":  # hostname is actually motd in the query protocol
                match = re.search(
                    b"(.*?)\x00(hostip|hostport|game_id|gametype|map|maxplayers|numplayers|plugins|version)",
                    response.received,
                    flags=re.DOTALL,
                )
                motd = match.group(1) if match else ""
                # Since the query protocol does not properly support unicode, the motd is still not resolved
                # correctly; however, this will avoid other parameter parsing errors.
                data[key] = response.read(len(motd)).decode("ISO-8859-1")
                response.read(1)  # ignore null byte
            elif len(key) == 0:
                response.read(1)
                break
            else:
                value = response.read_ascii()
                data[key] = value

        response.read(len("player_") + 2)

        players_list = []
        while True:
            player = response.read_ascii()
            if len(player) == 0:
                break
            players_list.append(player)

        return RawQueryResponse(**data), players_list


@final
@dataclass
class ServerQuerier(_BaseServerQuerier):
    connection: UDPSocketConnection  # pyright: ignore[reportIncompatibleVariableOverride]

    def _read_packet(self) -> Connection:
        packet = Connection()
        packet.receive(self.connection.read(self.connection.remaining()))
        packet.read(1 + 4)
        return packet

    def handshake(self) -> None:
        self.connection.write(self._create_handshake_packet())

        packet = self._read_packet()
        self.challenge = int(packet.read_ascii())

    def read_query(self) -> QueryResponse:
        request = self._create_packet()
        self.connection.write(request)

        response = self._read_packet()
        return QueryResponse.build(*self._parse_response(response))


@final
@dataclass
class AsyncServerQuerier(_BaseServerQuerier):
    connection: UDPAsyncSocketConnection  # pyright: ignore[reportIncompatibleVariableOverride]

    async def _read_packet(self) -> Connection:
        packet = Connection()
        packet.receive(await self.connection.read(self.connection.remaining()))
        packet.read(1 + 4)
        return packet

    async def handshake(self) -> None:
        await self.connection.write(self._create_handshake_packet())

        packet = await self._read_packet()
        self.challenge = int(packet.read_ascii())

    async def read_query(self) -> QueryResponse:
        request = self._create_packet()
        await self.connection.write(request)

        response = await self._read_packet()
        return QueryResponse.build(*self._parse_response(response))
